'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var neoBlockchainNetwork = require('neo-blockchain-network');
var neoBlockchainCore = require('neo-blockchain-core');
var neoBlockchainNodeCore = require('neo-blockchain-node-core');
var BloomFilter = _interopDefault(require('bloom-filter'));
var bloem = require('bloem');
var _ = _interopDefault(require('lodash'));
var net = _interopDefault(require('net'));
var BN = _interopDefault(require('bn.js'));
var stream = require('stream');

function _typeof(obj) {
  if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") {
    _typeof = function (obj) {
      return typeof obj;
    };
  } else {
    _typeof = function (obj) {
      return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
    };
  }

  return _typeof(obj);
}

function AwaitValue(value) {
  this.value = value;
}

function AsyncGenerator(gen) {
  var front, back;

  function send(key, arg) {
    return new Promise(function (resolve, reject) {
      var request = {
        key: key,
        arg: arg,
        resolve: resolve,
        reject: reject,
        next: null
      };

      if (back) {
        back = back.next = request;
      } else {
        front = back = request;
        resume(key, arg);
      }
    });
  }

  function resume(key, arg) {
    try {
      var result = gen[key](arg);
      var value = result.value;

      if (value instanceof AwaitValue) {
        Promise.resolve(value.value).then(function (arg) {
          resume("next", arg);
        }, function (arg) {
          resume("throw", arg);
        });
      } else {
        settle(result.done ? "return" : "normal", result.value);
      }
    } catch (err) {
      settle("throw", err);
    }
  }

  function settle(type, value) {
    switch (type) {
      case "return":
        front.resolve({
          value: value,
          done: true
        });
        break;

      case "throw":
        front.reject(value);
        break;

      default:
        front.resolve({
          value: value,
          done: false
        });
        break;
    }

    front = front.next;

    if (front) {
      resume(front.key, front.arg);
    } else {
      back = null;
    }
  }

  this._invoke = send;

  if (typeof gen.return !== "function") {
    this.return = undefined;
  }
}

if (typeof Symbol === "function" && Symbol.asyncIterator) {
  AsyncGenerator.prototype[Symbol.asyncIterator] = function () {
    return this;
  };
}

AsyncGenerator.prototype.next = function (arg) {
  return this._invoke("next", arg);
};

AsyncGenerator.prototype.throw = function (arg) {
  return this._invoke("throw", arg);
};

AsyncGenerator.prototype.return = function (arg) {
  return this._invoke("return", arg);
};

function _asyncToGenerator(fn) {
  return function () {
    var self = this,
        args = arguments;
    return new Promise(function (resolve, reject) {
      var gen = fn.apply(self, args);

      function step(key, arg) {
        try {
          var info = gen[key](arg);
          var value = info.value;
        } catch (error) {
          reject(error);
          return;
        }

        if (info.done) {
          resolve(value);
        } else {
          Promise.resolve(value).then(_next, _throw);
        }
      }

      function _next(value) {
        step("next", value);
      }

      function _throw(err) {
        step("throw", err);
      }

      _next();
    });
  };
}

function _classCallCheck(instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
}

function _defineProperties(target, props) {
  for (var i = 0; i < props.length; i++) {
    var descriptor = props[i];
    descriptor.enumerable = descriptor.enumerable || false;
    descriptor.configurable = true;
    if ("value" in descriptor) descriptor.writable = true;
    Object.defineProperty(target, descriptor.key, descriptor);
  }
}

function _createClass(Constructor, protoProps, staticProps) {
  if (protoProps) _defineProperties(Constructor.prototype, protoProps);
  if (staticProps) _defineProperties(Constructor, staticProps);
  return Constructor;
}

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

function _get(object, property, receiver) {
  if (object === null) object = Function.prototype;
  var desc = Object.getOwnPropertyDescriptor(object, property);

  if (desc === undefined) {
    var parent = Object.getPrototypeOf(object);

    if (parent === null) {
      return undefined;
    } else {
      return _get(parent, property, receiver);
    }
  } else if ("value" in desc) {
    return desc.value;
  } else {
    var getter = desc.get;

    if (getter === undefined) {
      return undefined;
    }

    return getter.call(receiver);
  }
}

function _inherits(subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function");
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      enumerable: false,
      writable: true,
      configurable: true
    }
  });
  if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass;
}

function _possibleConstructorReturn(self, call) {
  if (call && (typeof call === "object" || typeof call === "function")) {
    return call;
  }

  if (!self) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return self;
}

function _sliceIterator(arr, i) {
  var _arr = [];
  var _n = true;
  var _d = false;
  var _e = undefined;

  try {
    for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
      _arr.push(_s.value);

      if (i && _arr.length === i) break;
    }
  } catch (err) {
    _d = true;
    _e = err;
  } finally {
    try {
      if (!_n && _i["return"] != null) _i["return"]();
    } finally {
      if (_d) throw _e;
    }
  }

  return _arr;
}

function _slicedToArray(arr, i) {
  if (Array.isArray(arr)) {
    return arr;
  } else if (Symbol.iterator in Object(arr)) {
    return _sliceIterator(arr, i);
  } else {
    throw new TypeError("Invalid attempt to destructure non-iterable instance");
  }
}

var _fixBabelExtend = function (O) {
  var gOPD = O.getOwnPropertyDescriptor,
      gPO = O.getPrototypeOf || function (o) {
    return o.__proto__;
  },
      sPO = O.setPrototypeOf || function (o, p) {
    o.__proto__ = p;
    return o;
  },
      construct = _typeof(Reflect) === 'object' ? Reflect.construct : function (Parent, args, Class) {
    var Constructor,
        a = [null];
    a.push.apply(a, args);
    Constructor = Parent.bind.apply(Parent, a);
    return sPO(new Constructor(), Class.prototype);
  };

  return function fixBabelExtend(Class) {
    var Parent = gPO(Class);
    return sPO(Class, sPO(function Super() {
      return construct(Parent, arguments, gPO(this).constructor);
    }, Parent));
  };
}(Object);

var COMMAND = {
  ADDR: 'addr',
  BLOCK: 'block',
  CONSENSUS: 'consensus',
  FILTER_ADD: 'filteradd',
  FILTER_CLEAR: 'filterclear',
  FILTER_LOAD: 'filterload',
  GET_ADDR: 'getaddr',
  GET_BLOCKS: 'getblocks',
  GET_DATA: 'getdata',
  GET_HEADERS: 'getheaders',
  HEADERS: 'headers',
  INV: 'inv',
  MEMPOOL: 'mempool',
  TX: 'tx',
  VERACK: 'verack',
  VERSION: 'version',
  ALERT: 'alert',
  MERKLE_BLOCK: 'merkleblock',
  NOT_FOUND: 'notfound',
  PING: 'ping',
  PONG: 'pong',
  REJECT: 'reject'
};
var InvalidCommandError = _fixBabelExtend(
/*#__PURE__*/
function (_Error) {
  _inherits(InvalidCommandError, _Error);

  function InvalidCommandError(command) {
    var _this;

    _classCallCheck(this, InvalidCommandError);
    _this = _possibleConstructorReturn(this, (InvalidCommandError.__proto__ || Object.getPrototypeOf(InvalidCommandError)).call(this, "Expected command, found: ".concat(command)));
    _this.command = command;
    return _this;
  }

  return InvalidCommandError;
}(Error));
var assertCommand = function assertCommand(command) {
  switch (command) {
    case 'addr':
      return COMMAND.ADDR;

    case 'block':
      return COMMAND.BLOCK;

    case 'consensus':
      return COMMAND.CONSENSUS;

    case 'filteradd':
      return COMMAND.FILTER_ADD;

    case 'filterclear':
      return COMMAND.FILTER_CLEAR;

    case 'filterload':
      return COMMAND.FILTER_LOAD;

    case 'getaddr':
      return COMMAND.GET_ADDR;

    case 'getblocks':
      return COMMAND.GET_BLOCKS;

    case 'getdata':
      return COMMAND.GET_DATA;

    case 'getheaders':
      return COMMAND.GET_HEADERS;

    case 'headers':
      return COMMAND.HEADERS;

    case 'inv':
      return COMMAND.INV;

    case 'mempool':
      return COMMAND.MEMPOOL;

    case 'tx':
      return COMMAND.TX;

    case 'verack':
      return COMMAND.VERACK;

    case 'version':
      return COMMAND.VERSION;

    case 'alert':
      return COMMAND.ALERT;

    case 'merkleblock':
      return COMMAND.MERKLE_BLOCK;

    case 'notfound':
      return COMMAND.NOT_FOUND;

    case 'ping':
      return COMMAND.PING;

    case 'pong':
      return COMMAND.PONG;

    case 'reject':
      return COMMAND.REJECT;

    default:
      throw new InvalidCommandError(command);
  }
};

var NetworkAddress =
/*#__PURE__*/
function () {
  function NetworkAddress(_ref) {
    var host = _ref.host,
        port = _ref.port,
        timestamp = _ref.timestamp,
        services = _ref.services;
    _classCallCheck(this, NetworkAddress);
    Object.defineProperty(this, "serializeWire", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: neoBlockchainCore.createSerializeWire(this.serializeWireBase.bind(this))
    });
    this.host = host;
    this.port = port;
    this.timestamp = timestamp;
    this.services = services;
  }

  _createClass(NetworkAddress, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      // TODO: Support non-ipv4 addresses
      var parts = this.host.split('.');

      if (parts.length !== 4) {
        throw new neoBlockchainCore.InvalidFormatError();
      }

      var host = Buffer.from(parts.map(function (part) {
        return parseInt(part, 10);
      }));
      writer.writeUInt8(this.timestamp);
      writer.writeUInt64LE(this.services);
      writer.writeBytes(host); // TODO: NEO node reverses endianness of machine, so this is
      //       technically not well-defined. Let's assume it's always
      //       trying to write little endian values.

      writer.writeUInt16LE(this.port);
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(_ref2) {
      var reader = _ref2.reader;
      var timestamp = reader.readUInt8();
      var services = reader.readUInt64LE();
      var host0 = reader.readUInt8();
      var host1 = reader.readUInt8();
      var host2 = reader.readUInt8();
      var host3 = reader.readUInt8();
      var port = reader.readUInt16LE();
      return new this({
        timestamp: timestamp,
        services: services,
        host: [host0, host1, host2, host3].join('.'),
        port: port
      });
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new neoBlockchainCore.BinaryReader(options.buffer)
      });
    }
  }]);
  return NetworkAddress;
}();

var AddrPayload =
/*#__PURE__*/
function () {
  function AddrPayload(_ref) {
    var addresses = _ref.addresses;
    _classCallCheck(this, AddrPayload);
    Object.defineProperty(this, "serializeWire", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: neoBlockchainCore.createSerializeWire(this.serializeWireBase.bind(this))
    });
    this.addresses = addresses;
  }

  _createClass(AddrPayload, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      writer.writeArray(this.addresses, function (address) {
        return address.serializeWireBase(writer);
      });
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader;
      var addresses = reader.readArray(function () {
        return NetworkAddress.deserializeWireBase(options);
      });
      return new this({
        addresses: addresses
      });
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new neoBlockchainCore.BinaryReader(options.buffer)
      });
    }
  }]);
  return AddrPayload;
}();

var ConsensusPayload =
/*#__PURE__*/
function () {
  function ConsensusPayload(_ref) {
    var version = _ref.version,
        previousHash = _ref.previousHash,
        blockIndex = _ref.blockIndex,
        validatorIndex = _ref.validatorIndex,
        timestamp = _ref.timestamp,
        data = _ref.data,
        script = _ref.script;
    _classCallCheck(this, ConsensusPayload);
    Object.defineProperty(this, "serializeWire", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: neoBlockchainCore.createSerializeWire(this.serializeWireBase.bind(this))
    });
    this.version = version;
    this.previousHash = previousHash;
    this.blockIndex = blockIndex;
    this.validatorIndex = validatorIndex;
    this.timestamp = timestamp;
    this.data = data;
    this.script = script;
  }

  _createClass(ConsensusPayload, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      writer.writeUInt32LE(this.version);
      writer.writeUInt256(this.previousHash);
      writer.writeUInt32LE(this.blockIndex);
      writer.writeUInt16LE(this.validatorIndex);
      writer.writeUInt32LE(this.timestamp);
      writer.writeVarBytesLE(this.data);
      writer.writeUInt8(1);
      this.script.serializeWireBase(writer);
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader;
      var version = reader.readUInt32LE();
      var previousHash = reader.readUInt256();
      var blockIndex = reader.readUInt32LE();
      var validatorIndex = reader.readUInt16LE();
      var timestamp = reader.readUInt32LE();
      var data = reader.readVarBytesLE();

      if (reader.readUInt8() !== 1) {
        throw new neoBlockchainCore.InvalidFormatError();
      }

      var script = neoBlockchainCore.Witness.deserializeWireBase(options);
      return new this({
        version: version,
        previousHash: previousHash,
        blockIndex: blockIndex,
        validatorIndex: validatorIndex,
        timestamp: timestamp,
        data: data,
        script: script
      });
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new neoBlockchainCore.BinaryReader(options.buffer)
      });
    }
  }]);
  return ConsensusPayload;
}();

var FilterAddPayload =
/*#__PURE__*/
function () {
  function FilterAddPayload(_ref) {
    var data = _ref.data;
    _classCallCheck(this, FilterAddPayload);
    Object.defineProperty(this, "serializeWire", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: neoBlockchainCore.createSerializeWire(this.serializeWireBase.bind(this))
    });
    this.data = data;
  }

  _createClass(FilterAddPayload, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      writer.writeVarBytesLE(this.data);
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(_ref2) {
      var reader = _ref2.reader;
      var data = reader.readVarBytesLE(520);
      return new this({
        data: data
      });
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new neoBlockchainCore.BinaryReader(options.buffer)
      });
    }
  }]);
  return FilterAddPayload;
}();

var FilterLoadPayload =
/*#__PURE__*/
function () {
  function FilterLoadPayload(_ref) {
    var filter = _ref.filter,
        k = _ref.k,
        tweak = _ref.tweak;
    _classCallCheck(this, FilterLoadPayload);
    Object.defineProperty(this, "serializeWire", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: neoBlockchainCore.createSerializeWire(this.serializeWireBase.bind(this))
    });
    this.filter = filter;
    this.k = k;
    this.tweak = tweak;
  }

  _createClass(FilterLoadPayload, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      writer.writeVarBytesLE(this.filter);
      writer.writeUInt8(this.k);
      writer.writeUInt32LE(this.tweak);
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(_ref2) {
      var reader = _ref2.reader;
      var filter = reader.readVarBytesLE(36000);
      var k = reader.readUInt8();
      var tweak = reader.readUInt32LE();

      if (k > 50) {
        throw new neoBlockchainCore.InvalidFormatError();
      }

      return new this({
        filter: filter,
        k: k,
        tweak: tweak
      });
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new neoBlockchainCore.BinaryReader(options.buffer)
      });
    }
  }]);
  return FilterLoadPayload;
}();

var GetBlocksPayload =
/*#__PURE__*/
function () {
  function GetBlocksPayload(_ref) {
    var hashStart = _ref.hashStart,
        hashStop = _ref.hashStop;
    _classCallCheck(this, GetBlocksPayload);
    Object.defineProperty(this, "serializeWire", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: neoBlockchainCore.createSerializeWire(this.serializeWireBase.bind(this))
    });
    this.hashStart = hashStart;
    this.hashStop = hashStop || neoBlockchainCore.common.ZERO_UINT256;
  }

  _createClass(GetBlocksPayload, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      writer.writeArray(this.hashStart, function (value) {
        writer.writeUInt256(value);
      });
      writer.writeUInt256(this.hashStop);
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(_ref2) {
      var reader = _ref2.reader;
      var hashStart = reader.readArray(function () {
        return reader.readUInt256();
      });
      var hashStop = reader.readUInt256();
      return new this({
        hashStart: hashStart,
        hashStop: hashStop
      });
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new neoBlockchainCore.BinaryReader(options.buffer)
      });
    }
  }]);
  return GetBlocksPayload;
}();

var HeadersPayload =
/*#__PURE__*/
function () {
  function HeadersPayload(_ref) {
    var headers = _ref.headers;
    _classCallCheck(this, HeadersPayload);
    Object.defineProperty(this, "serializeWire", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: neoBlockchainCore.createSerializeWire(this.serializeWireBase.bind(this))
    });
    this.headers = headers;
  }

  _createClass(HeadersPayload, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      writer.writeArray(this.headers, function (header) {
        return header.serializeWireBase(writer);
      });
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader;
      var headers = reader.readArray(function () {
        return neoBlockchainCore.Header.deserializeWireBase(options);
      });
      return new this({
        headers: headers
      });
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new neoBlockchainCore.BinaryReader(options.buffer)
      });
    }
  }]);
  return HeadersPayload;
}();

var _fixBabelExtend$1 = function (O) {
  var gOPD = O.getOwnPropertyDescriptor,
      gPO = O.getPrototypeOf || function (o) {
    return o.__proto__;
  },
      sPO = O.setPrototypeOf || function (o, p) {
    o.__proto__ = p;
    return o;
  },
      construct = _typeof(Reflect) === 'object' ? Reflect.construct : function (Parent, args, Class) {
    var Constructor,
        a = [null];
    a.push.apply(a, args);
    Constructor = Parent.bind.apply(Parent, a);
    return sPO(new Constructor(), Class.prototype);
  };

  return function fixBabelExtend(Class) {
    var Parent = gPO(Class);
    return sPO(Class, sPO(function Super() {
      return construct(Parent, arguments, gPO(this).constructor);
    }, Parent));
  };
}(Object);

var INVENTORY_TYPE = {
  TRANSACTION: 0x01,
  BLOCK: 0x02,
  CONSENSUS: 0xe0
};
var InvalidInventoryTypeError = _fixBabelExtend$1(
/*#__PURE__*/
function (_Error) {
  _inherits(InvalidInventoryTypeError, _Error);

  function InvalidInventoryTypeError(inventoryType) {
    var _this;

    _classCallCheck(this, InvalidInventoryTypeError);
    _this = _possibleConstructorReturn(this, (InvalidInventoryTypeError.__proto__ || Object.getPrototypeOf(InvalidInventoryTypeError)).call(this, "Expected inventory type, found: ".concat(inventoryType)));
    _this.inventoryType = inventoryType;
    return _this;
  }

  return InvalidInventoryTypeError;
}(Error));
var assertInventoryType = function assertInventoryType(inventoryType) {
  switch (inventoryType) {
    case INVENTORY_TYPE.TRANSACTION:
      return INVENTORY_TYPE.TRANSACTION;

    case INVENTORY_TYPE.BLOCK:
      return INVENTORY_TYPE.BLOCK;

    case INVENTORY_TYPE.CONSENSUS:
      return INVENTORY_TYPE.CONSENSUS;

    default:
      throw new InvalidInventoryTypeError(inventoryType);
  }
};

var InvPayload =
/*#__PURE__*/
function () {
  function InvPayload(_ref) {
    var type = _ref.type,
        hashes = _ref.hashes;
    _classCallCheck(this, InvPayload);
    Object.defineProperty(this, "serializeWire", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: neoBlockchainCore.createSerializeWire(this.serializeWireBase.bind(this))
    });
    this.type = type;
    this.hashes = hashes;
  }

  _createClass(InvPayload, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      writer.writeUInt8(this.type);
      writer.writeArray(this.hashes, function (value) {
        writer.writeUInt256(value);
      });
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(_ref2) {
      var reader = _ref2.reader;
      var type = assertInventoryType(reader.readUInt8());
      var hashes = reader.readArray(function () {
        return reader.readUInt256();
      });
      return new this({
        type: type,
        hashes: hashes
      });
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new neoBlockchainCore.BinaryReader(options.buffer)
      });
    }
  }]);
  return InvPayload;
}();

/* @fl/*  */
var MerkleBlockPayload =
/*#__PURE__*/
function (_BlockBase) {
  _inherits(MerkleBlockPayload, _BlockBase);

  function MerkleBlockPayload(_ref) {
    var _this;

    var version = _ref.version,
        previousHash = _ref.previousHash,
        merkleRoot = _ref.merkleRoot,
        timestamp = _ref.timestamp,
        index = _ref.index,
        consensusData = _ref.consensusData,
        nextConsensus = _ref.nextConsensus,
        script = _ref.script,
        transactionCount = _ref.transactionCount,
        hashes = _ref.hashes,
        flags = _ref.flags;
    _classCallCheck(this, MerkleBlockPayload);
    _this = _possibleConstructorReturn(this, (MerkleBlockPayload.__proto__ || Object.getPrototypeOf(MerkleBlockPayload)).call(this, {
      version: version,
      previousHash: previousHash,
      merkleRoot: merkleRoot,
      timestamp: timestamp,
      index: index,
      consensusData: consensusData,
      nextConsensus: nextConsensus,
      script: script
    }));
    _this.transactionCount = transactionCount;
    _this.hashes = hashes;
    _this.flags = flags;
    _this.__merkleBlockPayloadSize = neoBlockchainCore.utils.lazy(function () {
      return _get(MerkleBlockPayload.prototype.__proto__ || Object.getPrototypeOf(MerkleBlockPayload.prototype), "size", _this) + neoBlockchainCore.IOHelper.sizeOfUInt32LE + neoBlockchainCore.IOHelper.sizeOfArray(_this.hashes, function () {
        return neoBlockchainCore.IOHelper.sizeOfUInt256;
      }) + neoBlockchainCore.IOHelper.sizeOfVarBytesLE(_this.flags);
    });
    return _this;
  }

  _createClass(MerkleBlockPayload, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      _get(MerkleBlockPayload.prototype.__proto__ || Object.getPrototypeOf(MerkleBlockPayload.prototype), "serializeWireBase", this).call(this, writer);
      writer.writeVarUIntLE(this.transactionCount);
      writer.writeArray(this.hashes, function (hash) {
        writer.writeUInt256(hash);
      });
      writer.writeVarBytesLE(this.flags);
    }
  }, {
    key: "size",
    get: function get() {
      return this.__merkleBlockPayloadSize();
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader;
      var blockBase = _get(MerkleBlockPayload.__proto__ || Object.getPrototypeOf(MerkleBlockPayload), "deserializeBlockBaseWireBase", this).call(this, options);
      var transactionCount = reader.readVarUIntLE(neoBlockchainCore.utils.INT_MAX_VALUE).toNumber();
      var hashes = reader.readArray(function () {
        return reader.readUInt256();
      });
      var flags = reader.readVarBytesLE();
      return new this({
        version: blockBase.version,
        previousHash: blockBase.previousHash,
        merkleRoot: blockBase.merkleRoot,
        timestamp: blockBase.timestamp,
        index: blockBase.index,
        consensusData: blockBase.consensusData,
        nextConsensus: blockBase.nextConsensus,
        script: blockBase.script,
        transactionCount: transactionCount,
        hashes: hashes,
        flags: flags
      });
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new neoBlockchainCore.BinaryReader(options.buffer)
      });
    }
  }]);
  return MerkleBlockPayload;
}(neoBlockchainCore.BlockBase);

var SERVICES = {
  NODE_NETWORK: new BN(1)
};

var VersionPayload =
/*#__PURE__*/
function () {
  function VersionPayload(_ref) {
    var protocolVersion = _ref.protocolVersion,
        services = _ref.services,
        timestamp = _ref.timestamp,
        port = _ref.port,
        nonce = _ref.nonce,
        userAgent = _ref.userAgent,
        startHeight = _ref.startHeight,
        relay = _ref.relay;
    _classCallCheck(this, VersionPayload);
    Object.defineProperty(this, "serializeWire", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: neoBlockchainCore.createSerializeWire(this.serializeWireBase.bind(this))
    });
    this.protocolVersion = protocolVersion;
    this.services = services;
    this.timestamp = timestamp;
    this.port = port;
    this.nonce = nonce;
    this.userAgent = userAgent;
    this.startHeight = startHeight;
    this.relay = relay;
  }

  _createClass(VersionPayload, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      writer.writeUInt32LE(this.protocolVersion);
      writer.writeUInt64LE(this.services);
      writer.writeUInt32LE(this.timestamp);
      writer.writeUInt16LE(this.port);
      writer.writeUInt32LE(this.nonce);
      writer.writeVarString(this.userAgent);
      writer.writeUInt32LE(this.startHeight);
      writer.writeBoolean(this.relay);
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(_ref2) {
      var reader = _ref2.reader;
      var protocolVersion = reader.readUInt32LE();
      var services = reader.readUInt64LE();
      var timestamp = reader.readUInt32LE();
      var port = reader.readUInt16LE();
      var nonce = reader.readUInt32LE();
      var userAgent = reader.readVarString(1024);
      var startHeight = reader.readUInt32LE();
      var relay = reader.readBoolean();
      return new this({
        protocolVersion: protocolVersion,
        services: services,
        timestamp: timestamp,
        port: port,
        nonce: nonce,
        userAgent: userAgent,
        startHeight: startHeight,
        relay: relay
      });
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new neoBlockchainCore.BinaryReader(options.buffer)
      });
    }
  }]);
  return VersionPayload;
}();

var _fixBabelExtend$2 = function (O) {
  var gOPD = O.getOwnPropertyDescriptor,
      gPO = O.getPrototypeOf || function (o) {
    return o.__proto__;
  },
      sPO = O.setPrototypeOf || function (o, p) {
    o.__proto__ = p;
    return o;
  },
      construct = _typeof(Reflect) === 'object' ? Reflect.construct : function (Parent, args, Class) {
    var Constructor,
        a = [null];
    a.push.apply(a, args);
    Constructor = Parent.bind.apply(Parent, a);
    return sPO(new Constructor(), Class.prototype);
  };

  return function fixBabelExtend(Class) {
    var Parent = gPO(Class);
    return sPO(Class, sPO(function Super() {
      return construct(Parent, arguments, gPO(this).constructor);
    }, Parent));
  };
}(Object);

var COMMAND_LENGTH = 12;
var PAYLOAD_MAX_SIZE = 0x02000000;

var calculateChecksum = function calculateChecksum(buffer) {
  return neoBlockchainCore.common.toUInt32LE(neoBlockchainCore.crypto.hash256(buffer));
};

var deserializeMessageHeader = function deserializeMessageHeader(_ref) {
  var context = _ref.context,
      reader = _ref.reader;

  if (reader.readUInt32LE() !== context.messageMagic) {
    throw new neoBlockchainCore.InvalidFormatError();
  }

  var command = assertCommand(reader.readFixedString(COMMAND_LENGTH));
  var length = reader.readUInt32LE();

  if (length > PAYLOAD_MAX_SIZE) {
    throw new neoBlockchainCore.InvalidFormatError();
  }

  var checksum = reader.readUInt32LE();
  return {
    command: command,
    length: length,
    checksum: checksum
  };
};

var Message =
/*#__PURE__*/
function () {
  function Message(_ref2) {
    var magic = _ref2.magic,
        value = _ref2.value;
    _classCallCheck(this, Message);
    Object.defineProperty(this, "serializeWire", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: neoBlockchainCore.createSerializeWire(this.serializeWireBase.bind(this))
    });
    this.magic = magic;
    this.value = value;
  }

  _createClass(Message, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      var value = this.value;
      writer.writeUInt32LE(this.magic);
      writer.writeFixedString(value.command, COMMAND_LENGTH);
      var payload = Buffer.alloc(0);

      switch (value.command) {
        case 'addr':
          payload = value.payload.serializeWire();
          break;

        case 'block':
          payload = value.payload.serializeWire();
          break;

        case 'consensus':
          payload = value.payload.serializeWire();
          break;

        case 'filteradd':
          payload = value.payload.serializeWire();
          break;

        case 'filterclear':
          break;

        case 'filterload':
          payload = value.payload.serializeWire();
          break;

        case 'getaddr':
          break;

        case 'getblocks':
          payload = value.payload.serializeWire();
          break;

        case 'getdata':
          payload = value.payload.serializeWire();
          break;

        case 'getheaders':
          payload = value.payload.serializeWire();
          break;

        case 'headers':
          payload = value.payload.serializeWire();
          break;

        case 'inv':
          payload = value.payload.serializeWire();
          break;

        case 'mempool':
          break;

        case 'tx':
          payload = value.payload.serializeWire();
          break;

        case 'verack':
          break;

        case 'version':
          payload = value.payload.serializeWire();
          break;

        case 'alert':
          break;

        case 'merkleblock':
          payload = value.payload.serializeWire();
          break;

        case 'notfound':
          break;

        case 'ping':
          break;

        case 'pong':
          break;

        case 'reject':
          break;

        default:
          // eslint-disable-next-line
          value.command;
          throw new neoBlockchainCore.InvalidFormatError();
      }

      writer.writeUInt32LE(payload.length);
      writer.writeUInt32LE(calculateChecksum(payload));
      writer.writeBytes(payload);
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader,
          context = options.context;

      var _deserializeMessageHe = deserializeMessageHeader(options),
          command = _deserializeMessageHe.command,
          length = _deserializeMessageHe.length,
          checksum = _deserializeMessageHe.checksum;

      var payloadBuffer = reader.readBytes(length);

      if (calculateChecksum(payloadBuffer) !== checksum) {
        throw new neoBlockchainCore.InvalidFormatError();
      }

      var payloadOptions = {
        context: options.context,
        buffer: payloadBuffer
      };
      var value;

      switch (command) {
        case 'addr':
          value = {
            command: COMMAND.ADDR,
            payload: AddrPayload.deserializeWire(payloadOptions)
          };
          break;

        case 'block':
          value = {
            command: COMMAND.BLOCK,
            payload: neoBlockchainCore.Block.deserializeWire(payloadOptions)
          };
          break;

        case 'consensus':
          value = {
            command: COMMAND.CONSENSUS,
            payload: ConsensusPayload.deserializeWire(payloadOptions)
          };
          break;

        case 'filteradd':
          value = {
            command: COMMAND.FILTER_ADD,
            payload: FilterAddPayload.deserializeWire(payloadOptions)
          };
          break;

        case 'filterclear':
          value = {
            command: COMMAND.FILTER_CLEAR
          };
          break;

        case 'filterload':
          value = {
            command: COMMAND.FILTER_LOAD,
            payload: FilterLoadPayload.deserializeWire(payloadOptions)
          };
          break;

        case 'getaddr':
          value = {
            command: COMMAND.GET_ADDR
          };
          break;

        case 'getblocks':
          value = {
            command: COMMAND.GET_BLOCKS,
            payload: GetBlocksPayload.deserializeWire(payloadOptions)
          };
          break;

        case 'getdata':
          value = {
            command: COMMAND.GET_DATA,
            payload: InvPayload.deserializeWire(payloadOptions)
          };
          break;

        case 'getheaders':
          value = {
            command: COMMAND.GET_HEADERS,
            payload: GetBlocksPayload.deserializeWire(payloadOptions)
          };
          break;

        case 'headers':
          value = {
            command: COMMAND.HEADERS,
            payload: HeadersPayload.deserializeWire(payloadOptions)
          };
          break;

        case 'inv':
          value = {
            command: COMMAND.INV,
            payload: InvPayload.deserializeWire(payloadOptions)
          };
          break;

        case 'mempool':
          value = {
            command: COMMAND.MEMPOOL
          };
          break;

        case 'tx':
          value = {
            command: COMMAND.TX,
            payload: neoBlockchainCore.deserializeTransactionWire(payloadOptions)
          };
          break;

        case 'verack':
          value = {
            command: COMMAND.VERACK
          };
          break;

        case 'version':
          value = {
            command: COMMAND.VERSION,
            payload: VersionPayload.deserializeWire(payloadOptions)
          };
          break;

        case 'alert':
          value = {
            command: COMMAND.ALERT
          };
          break;

        case 'merkleblock':
          value = {
            command: COMMAND.MERKLE_BLOCK,
            payload: MerkleBlockPayload.deserializeWire(payloadOptions)
          };
          break;

        case 'notfound':
          value = {
            command: COMMAND.NOT_FOUND
          };
          break;

        case 'ping':
          value = {
            command: COMMAND.PING
          };
          break;

        case 'pong':
          value = {
            command: COMMAND.PONG
          };
          break;

        case 'reject':
          value = {
            command: COMMAND.REJECT
          };
          break;

        default:
          // eslint-disable-next-line
          
          throw new neoBlockchainCore.InvalidFormatError();
      }

      return new this({
        magic: context.messageMagic,
        value: value
      });
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new neoBlockchainCore.BinaryReader(options.buffer)
      });
    }
  }]);
  return Message;
}();

var InvalidMessageTransformEncodingError = _fixBabelExtend$2(
/*#__PURE__*/
function (_Error) {
  _inherits(InvalidMessageTransformEncodingError, _Error);

  function InvalidMessageTransformEncodingError() {
    _classCallCheck(this, InvalidMessageTransformEncodingError);
    return _possibleConstructorReturn(this, (InvalidMessageTransformEncodingError.__proto__ || Object.getPrototypeOf(InvalidMessageTransformEncodingError)).call(this, 'Invalid Message Transform Encoding.'));
  }

  return InvalidMessageTransformEncodingError;
}(Error));
var SIZE_OF_MESSAGE_HEADER = neoBlockchainCore.IOHelper.sizeOfUInt32LE + neoBlockchainCore.IOHelper.sizeOfFixedString(COMMAND_LENGTH) + neoBlockchainCore.IOHelper.sizeOfUInt32LE + neoBlockchainCore.IOHelper.sizeOfUInt32LE;
var MessageTransform =
/*#__PURE__*/
function (_Transform) {
  _inherits(MessageTransform, _Transform);

  function MessageTransform(context) {
    var _this;

    _classCallCheck(this, MessageTransform);
    _this = _possibleConstructorReturn(this, (MessageTransform.__proto__ || Object.getPrototypeOf(MessageTransform)).call(this, {
      readableObjectMode: true
    }));
    _this.context = context;
    _this.buffer = Buffer.from([]);
    return _this;
  }

  _createClass(MessageTransform, [{
    key: "_transform",
    value: function _transform(chunk, encoding, callback) {
      var _this2 = this;

      if (typeof chunk === 'string' || encoding !== 'buffer') {
        throw new InvalidMessageTransformEncodingError();
      }

      this.buffer = Buffer.concat([this.buffer, chunk]);

      try {
        var _processBuffer2 = this._processBuffer(new neoBlockchainCore.BinaryReader(this.buffer)),
            remainingBuffer = _processBuffer2.remainingBuffer,
            messages = _processBuffer2.messages;

        this.buffer = remainingBuffer;
        messages.forEach(function (message) {
          return _this2.push(message);
        });
        callback(null);
      } catch (error) {
        callback(error);
      }
    }
  }, {
    key: "_processBuffer",
    value: function _processBuffer(reader) {
      if (reader.remaining < SIZE_OF_MESSAGE_HEADER) {
        return {
          remainingBuffer: reader.remainingBuffer,
          messages: []
        };
      }

      var _deserializeMessageHe2 = deserializeMessageHeader({
        context: this.context,
        reader: reader.clone()
      }),
          length = _deserializeMessageHe2.length;

      if (reader.remaining < SIZE_OF_MESSAGE_HEADER + length) {
        return {
          remainingBuffer: reader.remainingBuffer,
          messages: []
        };
      }

      var message = Message.deserializeWireBase({
        context: this.context,
        reader: reader
      });

      var _processBuffer3 = this._processBuffer(reader),
          remainingBuffer = _processBuffer3.remainingBuffer,
          messages = _processBuffer3.messages;

      messages.push(message);
      return {
        remainingBuffer: remainingBuffer,
        messages: messages
      };
    }
  }]);
  return MessageTransform;
}(stream.Transform);

var _fixBabelExtend$3 = function (O) {
  var gOPD = O.getOwnPropertyDescriptor,
      gPO = O.getPrototypeOf || function (o) {
    return o.__proto__;
  },
      sPO = O.setPrototypeOf || function (o, p) {
    o.__proto__ = p;
    return o;
  },
      construct = _typeof(Reflect) === 'object' ? Reflect.construct : function (Parent, args, Class) {
    var Constructor,
        a = [null];
    a.push.apply(a, args);
    Constructor = Parent.bind.apply(Parent, a);
    return sPO(new Constructor(), Class.prototype);
  };

  return function fixBabelExtend(Class) {
    var Parent = gPO(Class);
    return sPO(Class, sPO(function Super() {
      return construct(Parent, arguments, gPO(this).constructor);
    }, Parent));
  };
}(Object);

// eslint-disable-next-line
var NegotiationError = _fixBabelExtend$3(
/*#__PURE__*/
function (_Error) {
  _inherits(NegotiationError, _Error);

  function NegotiationError(message) {
    var _this;

    _classCallCheck(this, NegotiationError);
    _this = _possibleConstructorReturn(this, (NegotiationError.__proto__ || Object.getPrototypeOf(NegotiationError)).call(this, "Negotiation failed. Unexpected message received: ".concat(message.value.command)));
    _this.messageObj = message;
    return _this;
  }

  return NegotiationError;
}(Error));

var name = "neo-blockchain-node";
var version = "0.0.43";
var keywords = ["neo", "blockchain"];
var description = "A NEO Node implementation.";
var repository = "https://github.com/neotracker/neo-blockchain";
var bugs = "https://github.com/neotracker/neo-blockchain/issues";
var license = "MIT";
var main = "./dist/index.js";
var module$1 = "./dist/es.js";
var dependencies = {
  "bloem": "^0.2.4",
  "bloom-filter": "^0.2.0",
  "bn.js": "^4.11.8",
  "lodash": "^4.17.4",
  "neo-blockchain-core": "^0.0.43",
  "neo-blockchain-network": "^0.0.43",
  "neo-blockchain-node-core": "^0.0.43"
};
var pkg = {
  name: name,
  version: version,
  keywords: keywords,
  description: description,
  repository: repository,
  bugs: bugs,
  license: license,
  main: main,
  module: module$1,
  dependencies: dependencies
};

var createPeerBloomFilter = function createPeerBloomFilter(_ref) {
  var filter = _ref.filter,
      k = _ref.k,
      tweak = _ref.tweak;
  return new BloomFilter({
    vData: Buffer.from(filter),
    nHashFuncs: k,
    nTweak: tweak
  });
};

var createScalingBloomFilter = function createScalingBloomFilter() {
  return new bloem.ScalingBloem(0.05, {
    initial_capacity: 100000,
    scaling: 4
  });
};

var MEM_POOL_SIZE = 30000;
var GET_ADDR_PEER_COUNT = 200;
var GET_BLOCKS_COUNT = 500; // Assume that we get 500 back, but if not, at least request every 10 seconds

var GET_BLOCKS_BUFFER = GET_BLOCKS_COUNT / 3;
var GET_BLOCKS_TIME_MS = 10000;
var GET_BLOCKS_THROTTLE_MS = 500;
var GET_BLOCKS_CLOSE_COUNT = 3;
var LOCAL_HOST_ADDRESSES = new Set(['0.0.0.0', 'localhost', '127.0.0.1']);

var Node$1 =
/*#__PURE__*/
function () {
  function Node(options) {
    var _this = this;

    _classCallCheck(this, Node);
    Object.defineProperty(this, "_negotiate", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: function () {
        var _ref2 = _asyncToGenerator(
        /*#__PURE__*/
        regeneratorRuntime.mark(function _callee(peer) {
          var message, versionPayload, address, _getEndpointConfig, host, nextMessage;

          return regeneratorRuntime.wrap(function _callee$(_context) {
            while (1) {
              switch (_context.prev = _context.next) {
                case 0:
                  _this._sendMessage(peer, _this._createMessage({
                    command: COMMAND.VERSION,
                    payload: new VersionPayload({
                      protocolVersion: 0,
                      services: SERVICES.NODE_NETWORK,
                      timestamp: Math.round(Date.now() / 1000),
                      port: _this._externalPort,
                      nonce: _this._nonce,
                      userAgent: _this._userAgent,
                      startHeight: _this._blockchain.currentBlockIndex,
                      relay: true
                    })
                  }));

                  _context.next = 3;
                  return peer.receiveMessage(30000);

                case 3:
                  message = _context.sent;

                  if (!(message.value.command === COMMAND.VERSION)) {
                    _context.next = 8;
                    break;
                  }

                  versionPayload = message.value.payload;
                  _context.next = 9;
                  break;

                case 8:
                  throw new NegotiationError(message);

                case 9:
                  if (_this._checkVersion(versionPayload)) {
                    _context.next = 11;
                    break;
                  }

                  throw new NegotiationError(message);

                case 11:
                  _getEndpointConfig = neoBlockchainNodeCore.getEndpointConfig(peer.endpoint), host = _getEndpointConfig.host;

                  if (net.isIPv4(host)) {
                    address = new NetworkAddress({
                      host: host,
                      port: versionPayload.port,
                      timestamp: versionPayload.timestamp,
                      services: versionPayload.services
                    });
                  }

                  _this._sendMessage(peer, _this._createMessage({
                    command: COMMAND.VERACK
                  }));

                  _context.next = 16;
                  return peer.receiveMessage(30000);

                case 16:
                  nextMessage = _context.sent;

                  if (!(nextMessage.value.command !== COMMAND.VERACK)) {
                    _context.next = 19;
                    break;
                  }

                  throw new NegotiationError(nextMessage);

                case 19:
                  return _context.abrupt("return", {
                    data: {
                      nonce: versionPayload.nonce,
                      startHeight: versionPayload.startHeight,
                      bloomFilter: null,
                      address: address
                    },
                    relay: versionPayload.relay
                  });

                case 20:
                case "end":
                  return _context.stop();
              }
            }
          }, _callee, this);
        }));

        function value(_x) {
          return _ref2.apply(this, arguments);
        }

        return value;
      }()
    });
    Object.defineProperty(this, "_onEvent", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: function value(event) {
        if (event.event === 'PEER_CONNECT_SUCCESS') {
          var connectedPeer = event.extra.connectedPeer;

          if (_this._bestPeer == null || _this._bestPeer.data.startHeight < connectedPeer.data.startHeight) {
            _this._bestPeer = connectedPeer;

            _this._requestBlocks();
          }
        } else if (event.event === 'PEER_CLOSED') {
          if (_this._bestPeer === event.extra.peer) {
            _this._bestPeer = _this._findBestPeer();

            _this._requestBlocks();
          }
        }

        _this._blockchain.log(_extends({
          event: event.event
        }, event.data || {}));
      }
    });
    Object.defineProperty(this, "_requestBlocks", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: _.throttle(_.debounce(function () {
        var peer = _this._bestPeer;
        var block = _this._blockchain.currentBlock;

        if (peer != null && block.index < peer.data.startHeight) {
          if (_this._getBlocksRequestsCount > GET_BLOCKS_CLOSE_COUNT) {
            peer.close();
            _this._getBlocksRequestsCount = 0;
          } else if (_this._shouldRequestBlocks()) {
            if (_this._getBlocksRequestsIndex === block.index) {
              _this._getBlocksRequestsCount += 1;
            } else {
              _this._getBlocksRequestsCount = 1;
              _this._getBlocksRequestsIndex = block.index;
            }

            _this._getBlocksRequestTime = Date.now();

            _this._sendMessage(peer, _this._createMessage({
              command: COMMAND.GET_BLOCKS,
              payload: new GetBlocksPayload({
                hashStart: [block.hash]
              })
            }));
          }

          _this._requestBlocks();
        }
      }, GET_BLOCKS_THROTTLE_MS), GET_BLOCKS_THROTTLE_MS)
    });
    Object.defineProperty(this, "_onRequestEndpoints", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: function value() {
        _this._relay(_this._createMessage({
          command: COMMAND.GET_ADDR
        }));
      }
    });
    this._blockchain = options.blockchain;
    this._network = new neoBlockchainNetwork.Network({
      seeds: options.seeds,
      listenTCP: options.listenTCP,
      externalEndpoints: options.externalEndpoints,
      connectPeersDelayMS: options.connectPeersDelayMS,
      maxConnectedPeers: options.maxConnectedPeers,
      socketTimeoutMS: options.socketTimeoutMS,
      negotiate: this._negotiate,
      createMessageTransform: function createMessageTransform() {
        return new MessageTransform(_this._blockchain.deserializeWireContext);
      },
      onMessageReceived: this._onMessageReceived.bind(this),
      onRequestEndpoints: this._onRequestEndpoints.bind(this),
      onEvent: this._onEvent
    });
    this._started = false;
    this._stopped = false;
    this._externalPort = (options.listenTCP || {}).port || 0;
    this._nonce = Math.floor(Math.random() * neoBlockchainCore.utils.UINT_MAX_NUMBER);
    this._userAgent = "NEO:neo-blockchain-js:".concat(pkg.version);
    this.memPool = {};
    this._knownBlockHashes = createScalingBloomFilter();
    this._tempKnownBlockHashes = new Set();
    this._knownTransactionHashes = createScalingBloomFilter();
    this._tempKnownTransactionHashes = new Set();
    this._knownHeaderHashes = createScalingBloomFilter();
    this._tempKnownHeaderHashes = new Set();
    this._getBlocksRequestsIndex = null;
    this._getBlocksRequestTime = null;
    this._getBlocksRequestsCount = 1;
  }

  _createClass(Node, [{
    key: "start",
    value: function start() {
      if (this._started) {
        return;
      }

      this._started = true;
      this._stopped = false;

      this._blockchain.log({
        event: 'NODE_START'
      });

      try {
        this._network.start();
      } catch (error) {
        this._started = false;
        throw error;
      }
    }
  }, {
    key: "stop",
    value: function stop() {
      if (this._stopped) {
        return;
      }

      this._stopped = true;

      this._blockchain.log({
        event: 'NODE_STOP'
      });

      try {
        this._network.stop();

        this._started = false;
      } catch (error) {
        this._stopped = false;
        throw error;
      }
    }
  }, {
    key: "relayTransaction",
    value: function () {
      var _ref3 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee2(transaction) {
        var _hash, foundTransaction;

        return regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                if (!(transaction.type === neoBlockchainCore.TRANSACTION_TYPE.MINER || this.memPool[transaction.hashHex] != null || this._tempKnownTransactionHashes.has(transaction.hashHex))) {
                  _context2.next = 2;
                  break;
                }

                return _context2.abrupt("return");

              case 2:
                if (this._knownTransactionHashes.has(transaction.hash)) {
                  _context2.next = 28;
                  break;
                }

                _hash = neoBlockchainCore.common.uInt256ToString(transaction.hash);

                this._blockchain.log({
                  event: 'RELAY_TRANSACTION_START',
                  level: 'debug',
                  hash: _hash
                });

                this._tempKnownTransactionHashes.add(transaction.hashHex);

                _context2.prev = 6;
                _context2.next = 9;
                return this._blockchain.transaction.tryGet({
                  hash: transaction.hash
                });

              case 9:
                foundTransaction = _context2.sent;

                if (!(foundTransaction == null)) {
                  _context2.next = 19;
                  break;
                }

                _context2.next = 13;
                return this._blockchain.verifyTransaction({
                  transaction: transaction,
                  memPool: neoBlockchainCore.utils.values(this.memPool)
                });

              case 13:
                this.memPool[transaction.hashHex] = transaction;

                this._knownTransactionHashes.add(transaction.hash);

                this._relayTransaction(transaction);

                this._blockchain.log({
                  event: 'RELAY_TRANSACTION_SUCCESS',
                  level: 'debug',
                  hash: _hash
                });

                _context2.next = 19;
                return this._trimMemPool();

              case 19:
                _context2.next = 25;
                break;

              case 21:
                _context2.prev = 21;
                _context2.t0 = _context2["catch"](6);

                this._blockchain.log({
                  event: 'RELAY_TRANSACTION_ERROR',
                  hash: _hash,
                  error: _context2.t0
                });

                throw _context2.t0;

              case 25:
                _context2.prev = 25;

                this._tempKnownTransactionHashes.delete(transaction.hashHex);

                return _context2.finish(25);

              case 28:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2, this, [[6, 21, 25, 28]]);
      }));

      function relayTransaction(_x2) {
        return _ref3.apply(this, arguments);
      }

      return relayTransaction;
    }()
  }, {
    key: "_relay",
    value: function _relay(message) {
      this._blockchain.log({
        event: 'RELAY_MESSAGE',
        level: 'debug',
        command: message.value.command
      });

      this._network.relay(message.serializeWire());
    }
  }, {
    key: "_relayTransaction",
    value: function _relayTransaction(transaction) {
      var message = this._createMessage({
        command: COMMAND.INV,
        payload: new InvPayload({
          type: INVENTORY_TYPE.TRANSACTION,
          hashes: [transaction.hash]
        })
      });

      var messagePayload = message.serializeWire();
      var _iteratorNormalCompletion = true;
      var _didIteratorError = false;
      var _iteratorError = undefined;

      try {
        for (var _iterator = this._network.connectedPeers[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
          var _peer = _step.value;

          if (_peer.relay && this._testFilter(_peer.data.bloomFilter, transaction)) {
            _peer.write(messagePayload);
          }
        }
      } catch (err) {
        _didIteratorError = true;
        _iteratorError = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion && _iterator.return != null) {
            _iterator.return();
          }
        } finally {
          if (_didIteratorError) {
            throw _iteratorError;
          }
        }
      }
    }
  }, {
    key: "_sendMessage",
    value: function _sendMessage(peer, message) {
      this._blockchain.log({
        event: 'SEND_MESSAGE',
        level: 'debug',
        command: message.value.command,
        endpoint: peer.endpoint
      });

      peer.write(message.serializeWire());
    }
  }, {
    key: "_findBestPeer",
    value: function _findBestPeer() {
      return _.maxBy(this._network.connectedPeers, function (peer) {
        return peer.data.startHeight;
      });
    }
  }, {
    key: "_shouldRequestBlocks",
    value: function _shouldRequestBlocks() {
      var block = this._blockchain.currentBlock;
      var getBlocksRequestTime = this._getBlocksRequestTime;
      return this._getBlocksRequestsIndex == null || block.index - this._getBlocksRequestsIndex > GET_BLOCKS_BUFFER || getBlocksRequestTime == null || Date.now() - getBlocksRequestTime > GET_BLOCKS_TIME_MS;
    }
  }, {
    key: "_checkVersion",
    value: function _checkVersion(version$$1) {
      if (version$$1.nonce === this._nonce) {
        return false;
      }

      if (this._network.connectedPeers.some(function (peer) {
        return version$$1.nonce === peer.data.nonce;
      })) {
        return false;
      }

      return true;
    }
  }, {
    key: "_ready",
    value: function _ready() {
      var peer = this._bestPeer;
      var block = this._blockchain.currentBlock;
      return peer != null && block.index >= peer.data.startHeight;
    }
  }, {
    key: "_onMessageReceived",
    value: function _onMessageReceived(peer, message) {
      var _this2 = this;

      this._blockchain.log({
        event: 'MESSAGE_RECEIVED',
        level: 'debug',
        endpoint: peer.endpoint,
        command: message.value.command
      });

      var onError = function onError(error) {
        _this2._blockchain.log({
          event: 'MESSAGE_RECEIVED_ERROR',
          error: error,
          endpoint: peer.endpoint,
          command: message.value.command
        });
      };

      try {
        switch (message.value.command) {
          case COMMAND.ADDR:
            this._onAddrMessageReceived(message.value.payload);

            break;

          case COMMAND.BLOCK:
            this._onBlockMessageReceived(message.value.payload).catch(onError);

            break;

          case COMMAND.CONSENSUS:
            this._onConsensusMessageReceived(message.value.payload);

            break;

          case COMMAND.FILTER_ADD:
            this._onFilterAddMessageReceived(peer, message.value.payload);

            break;

          case COMMAND.FILTER_CLEAR:
            this._onFilterClearMessageReceived(peer);

            break;

          case COMMAND.FILTER_LOAD:
            this._onFilterLoadMessageReceived(peer, message.value.payload);

            break;

          case COMMAND.GET_ADDR:
            this._onGetAddrMessageReceived(peer);

            break;

          case COMMAND.GET_BLOCKS:
            this._onGetBlocksMessageReceived(peer, message.value.payload).catch(onError);

            break;

          case COMMAND.GET_DATA:
            this._onGetDataMessageReceived(peer, message.value.payload).catch(onError);

            break;

          case COMMAND.GET_HEADERS:
            this._onGetHeadersMessageReceived(peer, message.value.payload).catch(onError);

            break;

          case COMMAND.HEADERS:
            this._onHeadersMessageReceived(peer, message.value.payload).catch(onError);

            break;

          case COMMAND.INV:
            this._onInvMessageReceived(peer, message.value.payload);

            break;

          case COMMAND.MEMPOOL:
            this._onMemPoolMessageReceived(peer);

            break;

          case COMMAND.TX:
            this._onTransactionReceived(message.value.payload).catch(onError);

            break;

          case COMMAND.VERACK:
            this._onVerackMessageReceived(peer);

            break;

          case COMMAND.VERSION:
            this._onVersionMessageReceived(peer);

            break;

          case COMMAND.ALERT:
            break;

          case COMMAND.MERKLE_BLOCK:
            break;

          case COMMAND.NOT_FOUND:
            break;

          case COMMAND.PING:
            break;

          case COMMAND.PONG:
            break;

          case COMMAND.REJECT:
            break;

          default:
            // eslint-disable-next-line
            message.value.command;
            break;
        }
      } catch (error) {
        onError(error);
      }
    }
  }, {
    key: "_onAddrMessageReceived",
    value: function _onAddrMessageReceived(addr) {
      var _this3 = this;

      addr.addresses.filter(function (address) {
        return !LOCAL_HOST_ADDRESSES.has(address.host);
      }).map(function (address) {
        return neoBlockchainNodeCore.createEndpoint({
          type: 'tcp',
          host: address.host,
          port: address.port
        });
      }).forEach(function (endpoint) {
        return _this3._network.addEndpoint(endpoint);
      });
    }
  }, {
    key: "_onBlockMessageReceived",
    value: function () {
      var _ref4 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee3(block) {
        var foundBlock, peer, _iteratorNormalCompletion2, _didIteratorError2, _iteratorError2, _iterator2, _step2, _transaction;

        return regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                if (!(this._blockchain.currentBlockIndex >= block.index || this._tempKnownBlockHashes.has(block.hashHex))) {
                  _context3.next = 2;
                  break;
                }

                return _context3.abrupt("return");

              case 2:
                if (this._knownBlockHashes.has(block.hash)) {
                  _context3.next = 37;
                  break;
                }

                this._tempKnownBlockHashes.add(block.hashHex);

                _context3.prev = 4;
                _context3.next = 7;
                return this._blockchain.block.tryGet({
                  hashOrIndex: block.hash
                });

              case 7:
                foundBlock = _context3.sent;

                if (!(foundBlock == null)) {
                  _context3.next = 13;
                  break;
                }

                _context3.next = 11;
                return this._blockchain.persistBlock({
                  block: block
                });

              case 11:
                peer = this._bestPeer;

                if (peer != null && block.index > peer.data.startHeight) {
                  this._relay(this._createMessage({
                    command: COMMAND.INV,
                    payload: new InvPayload({
                      type: INVENTORY_TYPE.BLOCK,
                      hashes: [block.hash]
                    })
                  }));
                }

              case 13:
                this._knownBlockHashes.add(block.hash);

                this._knownHeaderHashes.add(block.hash);

                _iteratorNormalCompletion2 = true;
                _didIteratorError2 = false;
                _iteratorError2 = undefined;
                _context3.prev = 18;

                for (_iterator2 = block.transactions[Symbol.iterator](); !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
                  _transaction = _step2.value;
                  delete this.memPool[_transaction.hashHex];

                  this._knownTransactionHashes.add(_transaction.hash);
                }

                _context3.next = 26;
                break;

              case 22:
                _context3.prev = 22;
                _context3.t0 = _context3["catch"](18);
                _didIteratorError2 = true;
                _iteratorError2 = _context3.t0;

              case 26:
                _context3.prev = 26;
                _context3.prev = 27;

                if (!_iteratorNormalCompletion2 && _iterator2.return != null) {
                  _iterator2.return();
                }

              case 29:
                _context3.prev = 29;

                if (!_didIteratorError2) {
                  _context3.next = 32;
                  break;
                }

                throw _iteratorError2;

              case 32:
                return _context3.finish(29);

              case 33:
                return _context3.finish(26);

              case 34:
                _context3.prev = 34;

                this._tempKnownBlockHashes.delete(block.hashHex);

                return _context3.finish(34);

              case 37:
              case "end":
                return _context3.stop();
            }
          }
        }, _callee3, this, [[4,, 34, 37], [18, 22, 26, 34], [27,, 29, 33]]);
      }));

      function _onBlockMessageReceived(_x3) {
        return _ref4.apply(this, arguments);
      }

      return _onBlockMessageReceived;
    }() // TODO: Implement
    // eslint-disable-next-line

  }, {
    key: "_onConsensusMessageReceived",
    value: function _onConsensusMessageReceived(consensus) {} // eslint-disable-next-line

  }, {
    key: "_onFilterAddMessageReceived",
    value: function _onFilterAddMessageReceived(peer, filterAdd) {
      if (peer.data.bloomFilter != null) {
        peer.data.bloomFilter.insert(filterAdd.data);
      }
    } // eslint-disable-next-line

  }, {
    key: "_onFilterClearMessageReceived",
    value: function _onFilterClearMessageReceived(peer) {
      // eslint-disable-next-line
      peer.data.bloomFilter = null;
    } // eslint-disable-next-line

  }, {
    key: "_onFilterLoadMessageReceived",
    value: function _onFilterLoadMessageReceived(peer, filterLoad) {
      // eslint-disable-next-line
      peer.data.bloomFilter = createPeerBloomFilter(filterLoad);
    }
  }, {
    key: "_onGetAddrMessageReceived",
    value: function _onGetAddrMessageReceived(peer) {
      var addresses = _.take(_.shuffle(this._network.connectedPeers.map(function (connectedPeer) {
        return connectedPeer.data.address;
      }).filter(Boolean)), GET_ADDR_PEER_COUNT);

      if (addresses.length > 0) {
        this._sendMessage(peer, this._createMessage({
          command: COMMAND.ADDR,
          payload: new AddrPayload({
            addresses: addresses
          })
        }));
      }
    }
  }, {
    key: "_onGetBlocksMessageReceived",
    value: function () {
      var _ref5 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee4(peer, getBlocks) {
        var headers;
        return regeneratorRuntime.wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                _context4.next = 2;
                return this._getHeaders(getBlocks, this._blockchain.currentBlockIndex);

              case 2:
                headers = _context4.sent;

                this._sendMessage(peer, this._createMessage({
                  command: COMMAND.INV,
                  payload: new InvPayload({
                    type: INVENTORY_TYPE.BLOCK,
                    hashes: headers.map(function (header) {
                      return header.hash;
                    })
                  })
                }));

              case 4:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee4, this);
      }));

      function _onGetBlocksMessageReceived(_x4, _x5) {
        return _ref5.apply(this, arguments);
      }

      return _onGetBlocksMessageReceived;
    }()
  }, {
    key: "_onGetDataMessageReceived",
    value: function () {
      var _ref6 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee7(peer, getData) {
        var _this4 = this;

        return regeneratorRuntime.wrap(function _callee7$(_context7) {
          while (1) {
            switch (_context7.prev = _context7.next) {
              case 0:
                _context7.t0 = getData.type;
                _context7.next = _context7.t0 === 0x01 ? 3 : _context7.t0 === 0x02 ? 6 : _context7.t0 === 0xe0 ? 9 : 10;
                break;

              case 3:
                _context7.next = 5;
                return Promise.all(getData.hashes.map(function () {
                  var _ref7 = _asyncToGenerator(
                  /*#__PURE__*/
                  regeneratorRuntime.mark(function _callee5(hash) {
                    var transaction;
                    return regeneratorRuntime.wrap(function _callee5$(_context5) {
                      while (1) {
                        switch (_context5.prev = _context5.next) {
                          case 0:
                            transaction = _this4.memPool[neoBlockchainCore.common.uInt256ToHex(hash)];

                            if (!(transaction == null)) {
                              _context5.next = 5;
                              break;
                            }

                            _context5.next = 4;
                            return _this4._blockchain.transaction.tryGet({
                              hash: hash
                            });

                          case 4:
                            transaction = _context5.sent;

                          case 5:
                            if (transaction != null) {
                              _this4._sendMessage(peer, _this4._createMessage({
                                command: COMMAND.TX,
                                payload: transaction
                              }));
                            }

                          case 6:
                          case "end":
                            return _context5.stop();
                        }
                      }
                    }, _callee5, this);
                  }));

                  return function (_x8) {
                    return _ref7.apply(this, arguments);
                  };
                }()));

              case 5:
                return _context7.abrupt("break", 12);

              case 6:
                _context7.next = 8;
                return Promise.all(getData.hashes.map(function () {
                  var _ref8 = _asyncToGenerator(
                  /*#__PURE__*/
                  regeneratorRuntime.mark(function _callee6(hash) {
                    var block;
                    return regeneratorRuntime.wrap(function _callee6$(_context6) {
                      while (1) {
                        switch (_context6.prev = _context6.next) {
                          case 0:
                            _context6.next = 2;
                            return _this4._blockchain.block.tryGet({
                              hashOrIndex: hash
                            });

                          case 2:
                            block = _context6.sent;

                            if (block != null) {
                              if (peer.data.bloomFilter == null) {
                                _this4._sendMessage(peer, _this4._createMessage({
                                  command: COMMAND.BLOCK,
                                  payload: block
                                }));
                              } else {
                                _this4._sendMessage(peer, _this4._createMessage({
                                  command: COMMAND.MERKLE_BLOCK,
                                  payload: _this4._createMerkleBlockPayload({
                                    block: block,
                                    flags: block.transactions.map(function (transaction) {
                                      return _this4._testFilter(peer.data.bloomFilter, transaction);
                                    })
                                  })
                                }));
                              }
                            }

                          case 4:
                          case "end":
                            return _context6.stop();
                        }
                      }
                    }, _callee6, this);
                  }));

                  return function (_x9) {
                    return _ref8.apply(this, arguments);
                  };
                }()));

              case 8:
                return _context7.abrupt("break", 12);

              case 9:
                return _context7.abrupt("break", 12);

              case 10:
                // eslint-disable-next-line
                getData.type;
                return _context7.abrupt("break", 12);

              case 12:
              case "end":
                return _context7.stop();
            }
          }
        }, _callee7, this);
      }));

      function _onGetDataMessageReceived(_x6, _x7) {
        return _ref6.apply(this, arguments);
      }

      return _onGetDataMessageReceived;
    }()
  }, {
    key: "_onGetHeadersMessageReceived",
    value: function () {
      var _ref9 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee8(peer, getBlocks) {
        var headers;
        return regeneratorRuntime.wrap(function _callee8$(_context8) {
          while (1) {
            switch (_context8.prev = _context8.next) {
              case 0:
                _context8.next = 2;
                return this._getHeaders(getBlocks, this._blockchain.currentHeader.index);

              case 2:
                headers = _context8.sent;

                this._sendMessage(peer, this._createMessage({
                  command: COMMAND.HEADERS,
                  payload: new HeadersPayload({
                    headers: headers
                  })
                }));

              case 4:
              case "end":
                return _context8.stop();
            }
          }
        }, _callee8, this);
      }));

      function _onGetHeadersMessageReceived(_x10, _x11) {
        return _ref9.apply(this, arguments);
      }

      return _onGetHeadersMessageReceived;
    }()
  }, {
    key: "_onHeadersMessageReceived",
    value: function () {
      var _ref10 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee9(peer, headersPayload) {
        var _this5 = this;

        var headers, _iteratorNormalCompletion3, _didIteratorError3, _iteratorError3, _iterator3, _step3, _header5, _iteratorNormalCompletion4, _didIteratorError4, _iteratorError4, _iterator4, _step4, _header2, _iteratorNormalCompletion5, _didIteratorError5, _iteratorError5, _iterator5, _step5, _header4;

        return regeneratorRuntime.wrap(function _callee9$(_context9) {
          while (1) {
            switch (_context9.prev = _context9.next) {
              case 0:
                headers = headersPayload.headers.filter(function (header) {
                  return !_this5._knownHeaderHashes.has(header.hash) && !_this5._tempKnownHeaderHashes.has(header.hashHex);
                });

                if (!(headers.length > 0)) {
                  _context9.next = 64;
                  break;
                }

                _iteratorNormalCompletion3 = true;
                _didIteratorError3 = false;
                _iteratorError3 = undefined;
                _context9.prev = 5;

                for (_iterator3 = headers[Symbol.iterator](); !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
                  _header5 = _step3.value;

                  this._tempKnownHeaderHashes.add(_header5.hashHex);
                }

                _context9.next = 13;
                break;

              case 9:
                _context9.prev = 9;
                _context9.t0 = _context9["catch"](5);
                _didIteratorError3 = true;
                _iteratorError3 = _context9.t0;

              case 13:
                _context9.prev = 13;
                _context9.prev = 14;

                if (!_iteratorNormalCompletion3 && _iterator3.return != null) {
                  _iterator3.return();
                }

              case 16:
                _context9.prev = 16;

                if (!_didIteratorError3) {
                  _context9.next = 19;
                  break;
                }

                throw _iteratorError3;

              case 19:
                return _context9.finish(16);

              case 20:
                return _context9.finish(13);

              case 21:
                _context9.prev = 21;
                _context9.next = 24;
                return this._blockchain.persistHeaders(headers);

              case 24:
                _iteratorNormalCompletion4 = true;
                _didIteratorError4 = false;
                _iteratorError4 = undefined;
                _context9.prev = 27;

                for (_iterator4 = headers[Symbol.iterator](); !(_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done); _iteratorNormalCompletion4 = true) {
                  _header2 = _step4.value;

                  this._knownHeaderHashes.add(_header2.hash);
                }

                _context9.next = 35;
                break;

              case 31:
                _context9.prev = 31;
                _context9.t1 = _context9["catch"](27);
                _didIteratorError4 = true;
                _iteratorError4 = _context9.t1;

              case 35:
                _context9.prev = 35;
                _context9.prev = 36;

                if (!_iteratorNormalCompletion4 && _iterator4.return != null) {
                  _iterator4.return();
                }

              case 38:
                _context9.prev = 38;

                if (!_didIteratorError4) {
                  _context9.next = 41;
                  break;
                }

                throw _iteratorError4;

              case 41:
                return _context9.finish(38);

              case 42:
                return _context9.finish(35);

              case 43:
                _context9.prev = 43;
                _iteratorNormalCompletion5 = true;
                _didIteratorError5 = false;
                _iteratorError5 = undefined;
                _context9.prev = 47;

                for (_iterator5 = headers[Symbol.iterator](); !(_iteratorNormalCompletion5 = (_step5 = _iterator5.next()).done); _iteratorNormalCompletion5 = true) {
                  _header4 = _step5.value;

                  this._tempKnownHeaderHashes.delete(_header4.hashHex);
                }

                _context9.next = 55;
                break;

              case 51:
                _context9.prev = 51;
                _context9.t2 = _context9["catch"](47);
                _didIteratorError5 = true;
                _iteratorError5 = _context9.t2;

              case 55:
                _context9.prev = 55;
                _context9.prev = 56;

                if (!_iteratorNormalCompletion5 && _iterator5.return != null) {
                  _iterator5.return();
                }

              case 58:
                _context9.prev = 58;

                if (!_didIteratorError5) {
                  _context9.next = 61;
                  break;
                }

                throw _iteratorError5;

              case 61:
                return _context9.finish(58);

              case 62:
                return _context9.finish(55);

              case 63:
                return _context9.finish(43);

              case 64:
                if (this._blockchain.currentHeader.index < peer.data.startHeight) {
                  this._sendMessage(peer, this._createMessage({
                    command: COMMAND.GET_HEADERS,
                    payload: new GetBlocksPayload({
                      hashStart: [this._blockchain.currentHeader.hash]
                    })
                  }));
                }

              case 65:
              case "end":
                return _context9.stop();
            }
          }
        }, _callee9, this, [[5, 9, 13, 21], [14,, 16, 20], [21,, 43, 64], [27, 31, 35, 43], [36,, 38, 42], [47, 51, 55, 63], [56,, 58, 62]]);
      }));

      function _onHeadersMessageReceived(_x12, _x13) {
        return _ref10.apply(this, arguments);
      }

      return _onHeadersMessageReceived;
    }()
  }, {
    key: "_onInvMessageReceived",
    value: function _onInvMessageReceived(peer, inv) {
      var _this6 = this;

      var hashes;

      switch (inv.type) {
        case 0x01:
          // Transaction
          hashes = inv.hashes.filter(function (hash) {
            return !_this6._knownTransactionHashes.has(hash) && !_this6._tempKnownTransactionHashes.has(neoBlockchainCore.common.uInt256ToHex(hash));
          });
          break;

        case 0x02:
          // Block
          hashes = inv.hashes.filter(function (hash) {
            return !_this6._knownBlockHashes.has(hash) && !_this6._tempKnownBlockHashes.has(neoBlockchainCore.common.uInt256ToHex(hash));
          });
          break;

        case 0xe0:
          // Consensus
          // TODO: Implement
          hashes = [];
          break;

        default:
          // eslint-disable-next-line
          inv.type;
          hashes = [];
          break;
      }

      if (hashes.length > 0) {
        this._sendMessage(peer, this._createMessage({
          command: COMMAND.GET_DATA,
          payload: new InvPayload({
            type: inv.type,
            hashes: hashes
          })
        }));
      }
    }
  }, {
    key: "_onMemPoolMessageReceived",
    value: function _onMemPoolMessageReceived(peer) {
      this._sendMessage(peer, this._createMessage({
        command: COMMAND.INV,
        payload: new InvPayload({
          type: INVENTORY_TYPE.TRANSACTION,
          hashes: neoBlockchainCore.utils.values(this.memPool).map(function (transaction) {
            return transaction.hash;
          })
        })
      }));
    }
  }, {
    key: "_onTransactionReceived",
    value: function () {
      var _ref11 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee10(transaction) {
        return regeneratorRuntime.wrap(function _callee10$(_context10) {
          while (1) {
            switch (_context10.prev = _context10.next) {
              case 0:
                if (!this._ready()) {
                  _context10.next = 3;
                  break;
                }

                _context10.next = 3;
                return this.relayTransaction(transaction);

              case 3:
              case "end":
                return _context10.stop();
            }
          }
        }, _callee10, this);
      }));

      function _onTransactionReceived(_x14) {
        return _ref11.apply(this, arguments);
      }

      return _onTransactionReceived;
    }() // eslint-disable-next-line

  }, {
    key: "_onVerackMessageReceived",
    value: function _onVerackMessageReceived(peer) {
      peer.close();
    } // eslint-disable-next-line

  }, {
    key: "_onVersionMessageReceived",
    value: function _onVersionMessageReceived(peer) {
      peer.close();
    }
  }, {
    key: "_getHeaders",
    value: function () {
      var _ref12 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee11(getBlocks, maxHeight) {
        var _this7 = this;

        var hashStopIndexPromise, _ref13, _ref14, hashStartHeaders, hashEnd, hashStartHeader, hashStart, headers;

        return regeneratorRuntime.wrap(function _callee11$(_context11) {
          while (1) {
            switch (_context11.prev = _context11.next) {
              case 0:
                hashStopIndexPromise = Promise.resolve(maxHeight);

                if (getBlocks.hashStop !== 0) {
                  hashStopIndexPromise = this._blockchain.header.tryGet({
                    hashOrIndex: getBlocks.hashStop
                  }).then(function (hashStopHeader) {
                    return hashStopHeader == null ? maxHeight : Math.min(hashStopHeader.index, maxHeight);
                  });
                }

                _context11.next = 4;
                return Promise.all([Promise.all(getBlocks.hashStart.map(function (hash) {
                  return _this7._blockchain.header.tryGet({
                    hashOrIndex: hash
                  });
                })), hashStopIndexPromise]);

              case 4:
                _ref13 = _context11.sent;
                _ref14 = _slicedToArray(_ref13, 2);
                hashStartHeaders = _ref14[0];
                hashEnd = _ref14[1];
                hashStartHeader = _.head(_.orderBy(hashStartHeaders.filter(Boolean), [function (header) {
                  return header.index;
                }]));

                if (!(hashStartHeader == null)) {
                  _context11.next = 11;
                  break;
                }

                return _context11.abrupt("return", []);

              case 11:
                hashStart = hashStartHeader.index + 1;

                if (!(hashStart > maxHeight)) {
                  _context11.next = 14;
                  break;
                }

                return _context11.abrupt("return", []);

              case 14:
                _context11.next = 16;
                return Promise.all(_.range(hashStart, Math.min(hashStart + GET_BLOCKS_COUNT, hashEnd)).map(function (index) {
                  return _this7._blockchain.header.get({
                    hashOrIndex: index
                  });
                }));

              case 16:
                headers = _context11.sent;
                return _context11.abrupt("return", headers);

              case 18:
              case "end":
                return _context11.stop();
            }
          }
        }, _callee11, this);
      }));

      function _getHeaders(_x15, _x16) {
        return _ref12.apply(this, arguments);
      }

      return _getHeaders;
    }()
  }, {
    key: "_trimMemPool",
    value: function () {
      var _ref15 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee13() {
        var _this8 = this;

        var memPool, transactionAndFee, hashesToRemove, _iteratorNormalCompletion6, _didIteratorError6, _iteratorError6, _iterator6, _step6, _hash3;

        return regeneratorRuntime.wrap(function _callee13$(_context13) {
          while (1) {
            switch (_context13.prev = _context13.next) {
              case 0:
                memPool = neoBlockchainCore.utils.values(this.memPool);

                if (!(memPool.length > MEM_POOL_SIZE)) {
                  _context13.next = 25;
                  break;
                }

                _context13.next = 4;
                return Promise.all(memPool.map(function () {
                  var _ref16 = _asyncToGenerator(
                  /*#__PURE__*/
                  regeneratorRuntime.mark(function _callee12(transaction) {
                    var networkFee;
                    return regeneratorRuntime.wrap(function _callee12$(_context12) {
                      while (1) {
                        switch (_context12.prev = _context12.next) {
                          case 0:
                            _context12.next = 2;
                            return transaction.getNetworkFee({
                              getOutput: _this8._blockchain.output.get,
                              governingToken: _this8._blockchain.settings.governingToken,
                              utilityToken: _this8._blockchain.settings.utilityToken,
                              fees: _this8._blockchain.settings.fees
                            });

                          case 2:
                            networkFee = _context12.sent;
                            return _context12.abrupt("return", [transaction, networkFee]);

                          case 4:
                          case "end":
                            return _context12.stop();
                        }
                      }
                    }, _callee12, this);
                  }));

                  return function (_x17) {
                    return _ref16.apply(this, arguments);
                  };
                }()));

              case 4:
                transactionAndFee = _context13.sent;
                hashesToRemove = _.take(_.sortBy(transactionAndFee, // TODO: Might be a bug since we're converting to number
                function (_ref17) {
                  var _ref18 = _slicedToArray(_ref17, 2),
                      transaction = _ref18[0],
                      networkFee = _ref18[1];

                  return networkFee.divn(transaction.size).toNumber();
                }) // eslint-disable-next-line
                ).map(function (_ref19) {
                  var _ref20 = _slicedToArray(_ref19, 2),
                      transaction = _ref20[0];

                  return transaction.hashHex;
                });
                _iteratorNormalCompletion6 = true;
                _didIteratorError6 = false;
                _iteratorError6 = undefined;
                _context13.prev = 9;

                for (_iterator6 = hashesToRemove[Symbol.iterator](); !(_iteratorNormalCompletion6 = (_step6 = _iterator6.next()).done); _iteratorNormalCompletion6 = true) {
                  _hash3 = _step6.value;
                  delete this.memPool[_hash3];
                }

                _context13.next = 17;
                break;

              case 13:
                _context13.prev = 13;
                _context13.t0 = _context13["catch"](9);
                _didIteratorError6 = true;
                _iteratorError6 = _context13.t0;

              case 17:
                _context13.prev = 17;
                _context13.prev = 18;

                if (!_iteratorNormalCompletion6 && _iterator6.return != null) {
                  _iterator6.return();
                }

              case 20:
                _context13.prev = 20;

                if (!_didIteratorError6) {
                  _context13.next = 23;
                  break;
                }

                throw _iteratorError6;

              case 23:
                return _context13.finish(20);

              case 24:
                return _context13.finish(17);

              case 25:
              case "end":
                return _context13.stop();
            }
          }
        }, _callee13, this, [[9, 13, 17, 25], [18,, 20, 24]]);
      }));

      function _trimMemPool() {
        return _ref15.apply(this, arguments);
      }

      return _trimMemPool;
    }()
  }, {
    key: "_testFilter",
    value: function _testFilter(bloomFilterIn, transaction) {
      var bloomFilter = bloomFilterIn;

      if (bloomFilter == null) {
        return true;
      }

      return bloomFilter.contains(transaction.hash) || transaction.outputs.some(function (output) {
        return bloomFilter.contains(output.address);
      }) || transaction.inputs.some(function (input) {
        return bloomFilter.contains(input.serializeWire());
      }) || transaction.scripts.some(function (script) {
        return bloomFilter.contains(neoBlockchainCore.crypto.toScriptHash(script.verification));
      }) || transaction.type === neoBlockchainCore.TRANSACTION_TYPE.REGISTER && transaction instanceof neoBlockchainCore.RegisterTransaction && bloomFilter.contains(transaction.asset.admin);
    }
  }, {
    key: "_createMerkleBlockPayload",
    value: function _createMerkleBlockPayload(_ref21) {
      var block = _ref21.block,
          flags = _ref21.flags;
      var tree = new neoBlockchainCore.MerkleTree(block.transactions.map(function (transaction) {
        return transaction.hash;
      })).trim(flags);
      var buffer = Buffer.allocUnsafe(Math.floor((flags.length + 7) / 8));

      for (var i = 0; i < flags.length; i += 1) {
        if (flags[i]) {
          // eslint-disable-next-line
          buffer[Math.floor(i / 8)] |= 1 << i % 8;
        }
      }

      return new MerkleBlockPayload({
        version: block.version,
        previousHash: block.previousHash,
        merkleRoot: block.merkleRoot,
        timestamp: block.timestamp,
        index: block.index,
        consensusData: block.consensusData,
        nextConsensus: block.nextConsensus,
        script: block.script,
        transactionCount: block.transactions.length,
        hashes: tree.toHashArray(),
        flags: buffer
      });
    }
  }, {
    key: "_createMessage",
    value: function _createMessage(value) {
      return new Message({
        magic: this._blockchain.settings.messageMagic,
        value: value
      });
    }
  }, {
    key: "connectedPeersCount",
    get: function get() {
      return this._network.connectedPeers.length;
    }
  }]);
  return Node;
}();

module.exports = Node$1;
//# sourceMappingURL=index.js.map
